import random, sys

# McmcParam types
CONTINUOUS = 1
DISCRETE = 2
BINARY = 4


class McmcParam:
  """
  Define a parameter to be sampled during an MCMC run.
  
  This is a fairly flexible class that the Mcmc class knows how to use for sampling.
  """
  def __init__(self, name, minval, maxval, startval, step, paramtype=CONTINUOUS):
    """Construct an McmcParam object.
    
    Parameters:
    name:      a name for the parameter (not used at present)
    minval:    the minimum value the parameter can take (smaller values will be
               rejected)
    maxval:    the maximum value the parameter can take (larger values will be
               rejected)
    startval:  the initial value for the parameter
    step:      maximum step size for values (for discrete parameters, fixed
               step size)
    paramtype: defines how the paramter steps. Should be mcmctools.CONTINUOUS 
               (the default), mcmctools.DISCRETE, or mcmctools.BINARY
    """
    
    self.name = name
    self.minval = minval
    self.maxval = maxval
    self.step = step
    self.currentval = startval
    self.lastval = startval
    self.paramtype = paramtype
    
  def stepback(self):
    """Step back to previous value."""
    
    self.currentval = self.lastval
    
  def stepforward(self):
    """Sample a new value value.
    
    Return: true iff sampled value is within the valid range.
    """
    self.lastval = self.currentval
    if self.paramtype == CONTINUOUS:
      newval = self.currentval + ((random.random() - 0.5) * 2 * self.step)
      self.currentval = newval
      
    elif self.paramtype == DISCRETE:
      if random.random() > 0.5:
        self.currentval += self.step
      else:
        self.currentval -= self.step
        
    else:
      self.currentval = not self.currentval
        
    return self.isvalid()
    
  def isvalid(self):
    """Determine whether the current value is within the valid range.
    
    Return:  true iff current value is valid
    """
    if self.currentval > self.maxval or self.currentval < self.minval:
      return False
    
    return True
  
  def getval(self):
    """Access the parameter's current value.
    
    Return: the current value 
    """
    return self.currentval
  
class Mcmc:
  """
  Define an Mcmc run. 
  
  This class is normally used to create an object of type Mcmc on which
  the run method is then called to start the run.
  """
  def __init__(self, paramdict, nsteps, kwargs, simulatemethod, acceptmethod, logfile=sys.stdout, startstep=0, saveevery=1):
    """Construct an Mcmc object.
    
    Parameters:
    paramdict:      a dictionary containing objects of type McmcParam with keys
                    (strings) that are used in the log file
    nsteps:         number of steps to run the MCMC for
    kwargs:         a dictionary that is passed to the simulation function (not
                    directly used here). If you need to know what the current step
                    number is you should include an integer with the key "stepcount"
    simulatemethod: method called to do actual simulation work
    acceptmethod:   method called to determine whether simulation was "successful"
    logfile:        open file to which sampled parameters are logged
    startstep:      number to assign to first simulation. This can be useful for 
                    resuming a simulation. (default = 0)
    saveevery:      frequency to sample and save parameter values
    """
    
    self.paramdict = paramdict
    self.nsteps = nsteps
    self.kwargs = kwargs
    self.simulate = simulatemethod
    self.accept = acceptmethod
    self.logfile = logfile
    self.startstep = startstep
    self.saveevery = saveevery
    
    if 'stepcount' in self.kwargs:
      self.foundcount = True
    else:
      self.foundcount = False
      
    
  def run(self):
    """Begin the MCMC run."""
    pkeys = self.paramdict.keys()
    pkeys.sort()
    
    if self.startstep == 0 and pkeys:
        self.logfile.write('samp')
      
        for k in pkeys:
          self.logfile.write('\t%s' % k)
        self.logfile.write('\n')
    
    if self.foundcount:
      self.kwargs['stepcount'] = startstep
    
    for i in xrange(self.startstep, self.nsteps):
      if self.foundcount:
        self.kwargs['stepcount'] += 1
        
      if not i % self.saveevery:
        self.logfile.write('%d' % (i + 1))
        for k in pkeys:
          self.logfile.write('\t%s' % self.paramdict[k].getval())
        self.logfile.write('\n')
        self.logfile.flush()
      
      pkey = random.sample(self.paramdict, 1)[0]
      
      doaccept = self.paramdict[pkey].stepforward()
      
      if doaccept:        
        self.simulate(self.paramdict, self.kwargs)
        doaccept = self.accept(self.kwargs)
      
      if not doaccept:
        self.paramdict[pkey].stepback()      
  
    
class McmcError(Exception):
  """Define an Exception for MCMC-related errors."""
  def __init__(self, msg):
    self.msg = msg
    
  def __repr__(self):
    return self.msg
  
# If you run this module directly, you'll do a UPGMA vs. NJ MCMC simulation. Fair warning.
if __name__ == '__main__':

  #some constants to play with
  ntaxa = 30

  samples = 100000
  saveevery = 100
  didelotsamps = 1000

  try:
    import math
    import p4, dendropy
    import numpy
    from scipy import stats
    import distance, disttree
    from string import letters
    from dendropy import treesim
    
  except ImportError:
    sys.stderr.write('One or more required packages could not be imported. If you wish to import\n')
    sys.stderr.write('mcmctools, you only need to have a working Python 2 installation. If you wish to\n')
    sys.stderr.write('run a UPGMA vs. NJ MCMC simulation, you need the following packages installed:\n\n')

    sys.stderr.write('  DendroPy: http://packages.python.org/DendroPy/\n')
    sys.stderr.write('  P4: http://www.bmnh.org/web_users/pf/p4.html\n')
    sys.stderr.write('  SciPy: http://www.scipy.org/\n')
    sys.stderr.write('  NumPy: http://numpy.scipy.org/\n')
    sys.stderr.write('  distree.py and distance.py, which are distributed with this module.\n\n')
    sys.exit(1)
     
  def calcheightREC(node, root):
    """Recursively calculate the height of the tree (starting at node)"""
    if node._child_nodes:
      heights = []
      for c in node._child_nodes:
        heights.append(calcheightREC(c, root))
      if node == root:
        return max(heights)
      else:
        return max(heights) + node.edge.length
    else:
      return node.edge.length

  def setbrlens(tree, scale, skew):
    """Randomly set edge lengths using scale and skew parameters.
    
    Scale is just the new height of the tree. Skew affects how clocklike the tree
    will be. Each edge is initially set by randomly drawing a uniform number u between
    +/- skew, and multiplying the edge by exp{u}.
    
    Parameters:
    scale: height of the tree
    skew:  "non-clocklikeness" of the tree
    """
    
    for e in tree.postorder_edge_iter():
      if e.length != None:
        u = 2 * skew * (random.random() - 0.5)   
        e.length *= math.exp(u)#(scale * math.exp(u))
    
    # set total tree height to scale
    oldheight = calcheightREC(tree.seed_node, tree.seed_node)
    factor = float(scale)/oldheight
    
    for e in tree.postorder_edge_iter():
      if e.length != None:
        e.length *= factor
    


  def yulesim(paramdict, kwargs):
    """Run a single simulation.
    
    A tree is generated under the Yule model, then edge lengths are adjusted according
    to scale and skew values (McmcParams with keys 'scale' and 'skew' in paramdict).
    Sequences are generated, distances computed, and trees estimated using NJ and UPGMA.
    Distances are then computed between the inferred trees and the true trees and stored
    in kwargs.
    """
    
    tree = treesim.pop_gen_tree(taxon_set=kwargs['taxa'])
    setbrlens(tree, paramdict['scale'].getval(), paramdict['skew'].getval())
    
    p4.read(str(tree).rsplit(':', 1)[0] + ';')
    tree_p4 = p4.var.trees[0]

    tree_p4.data = p4.Data(kwargs['align'])
    tree_p4.setPInvar(free=0, val=0)
    tree_p4.newComp(free=0, spec='specified', val=kwargs['comp'])   
    tree_p4.newRMatrix(free=0, spec='specified', val=kwargs['rmat'])
      
    tree_p4.simulate()
    dmat = distance.jc(tree_p4.data.alignments[0])
    upgmatree = disttree.upgma(dmat)
    utstr = disttree.treelist2string((upgmatree,), kwargs['taxa'])
    njtree = disttree.nj(dmat)
    ntstr = disttree.treelist2string((njtree,), kwargs['taxa'])
        
    p4.read(utstr)
    p4.var.trees[1].taxNames = tree_p4.taxNames
    p4.read(ntstr)
    p4.var.trees[2].taxNames = tree_p4.taxNames
      
    kwargs['upgma'] = tree_p4.topologyDistance(p4.var.trees[1])
    kwargs['nj'] = tree_p4.topologyDistance(p4.var.trees[2])
      
    
    del p4.var.trees[2]
    del p4.var.trees[1]
    del p4.var.trees[0]
    

  def yuleaccept(kwargs):
    """Determine whether the last simulation was successful.
    
    Return: true iff the UPGMA tree was at least as close to the true 
            tree as the NJ tree
    """
    return kwargs['upgma'] <= kwargs['nj']
    
  def readsamplefile(filename):
    """Read parameter values from a sample file.
    
    Parameters:
    filename: name of the file to read samples from
    
    Return: Tuple containing samples as a list of lists (one set per nested list)
            and names of parameters read from header line
    """
    
    f = open(filename)
    names = f.readline().split()[1:]
    
    samples = []
    
    for line in f:
      samples.append([float(val) for val in line.split()[1:]])
      
    f.close()
    
    return samples, names


  # Simulation code starts here

  if ntaxa > len(letters):
    sys.stderr.write('Too many taxa for me, sorry!\n')
    
  p4.var.warnReadNoFile = False

  # Jukes-Cantor for simulation
  comp = [0.25, 0.25, 0.25]
  rmat = [1.0/6 for i in xrange(6)]
  taxa = dendropy.TaxonSet()
  
  letters = letters[:26]
  for i in xrange(ntaxa):
    if i >= len(letters):
      label = '%s%d' % (letters[i % len(letters)], i / len(letters))
    else:
      if ntaxa > len(letters):
        label = '%s0' % letters[i]
      else:
        label = letters[i]
    taxa.new_taxon(label=label)
        
  a = p4.func.newEmptyAlignment(dataType='dna', taxNames=[str(t) for t in taxa], length=1000)
  
  kwargs = {'taxa':taxa, 'comp':comp, 'rmat':rmat, 'align':a}
  
  
  skew = McmcParam('skew', 0.0, math.log(10), 0.0, 0.5)
  scale = McmcParam('scale', 0.02, 1.0, 0.5, 2.0)
  
  paramdict = {'skew':skew, 'scale':scale}
  
  
  
  # Continue an interrupted run, checking for log file:
  nextstep = 0
  try:
    logfile = open('upgmanj.out', 'r')
    header = logfile.readline()
    keys = header.split()
    
    for line in logfile:
      pass
    
    lastentry = line.split()
    for i in xrange(len(lastentry)):
      k = keys[i]
      
      if k == 'samp':
        nextstep = int(lastentry[i]) + 1
      else:
        try:
          param = paramdict[k]
          if param.paramtype == CONTINUOUS:
            param.currentval = float(lastentry[i])
          elif param.paramtype == DISCRETE:
            param.currentval = int(lastentry[i])
          elif param.paramtype == BINARY:
            param.currentval = bool(lastentry[i])           
          else:
            sys.stderr.write('Undefined parameter type: %s\n' % str(param.paramtype))
            sys.exit(1)
            
        except KeyError:
          sys.stderr.write('Values in log file do not appear to correspond to this run. Exiting.\n')
          sys.exit(1)
          
    logfile.close()
          
  except IOError:
    sys.stderr.write('Log file not found. Starting a new run.\n')
  
  logfile = open('upgmanj.out', 'a')
    
  mcmcobj = Mcmc(paramdict, samples, kwargs, yulesim, yuleaccept, logfile, saveevery=saveevery, startstep=nextstep)
  
  mcmcobj.run()
  logfile.close()
  
  # Example of estimating P(success) using Didelot's (2011) method
  
  data, keys = readsamplefile('upgmanj.out')
  nparams = len(keys)
  niter = len(data)
  keepevery = 1 # You should look at ESS to determine what this should be 

  datamat = numpy.empty(dtype='float', shape=(nparams, niter / keepevery))
  
  for i in xrange(0, niter, keepevery):
    try:
      datamat[:, i/keepevery] = data[i]
    except IndexError:
      sys.stderr.write('Weird error (that I\'ll try to ignore) sampling from parameter values.\n')
      break
      
  kde = stats.gaussian_kde(datamat)
  sample = kde.resample(didelotsamps)
  weightsfile = open('weights.out','w')

  W = []
  paramlist = [paramdict[param] for param in keys]
  p_Theta = 1.0 #/(theta.maxval - theta.minval)
  
  for param in keys:
    p_Theta /= (paramdict[param].maxval - paramdict[param].minval)
  
  for i in xrange(didelotsamps):
    if i and not i % 100: 
      print 'Starting sample %d...' % (i + 1)
    doaccept = True
    for j in xrange(len(paramlist)):
      paramlist[j].lastval = paramlist[j].currentval
      paramlist[j].currentval = sample[j,i]
      
      if not paramlist[j].isvalid():
        doaccept = False
        
    w_i = 0.
    q_Theta_i = kde(sample[:,i])[0]
    
    if doaccept:
      yulesim(paramdict, kwargs)
      doaccept = yuleaccept(kwargs)
  
    if doaccept:
      w_i = p_Theta / q_Theta_i
      
    weightsfile.write('%g\n' % w_i)
    weightsfile.flush()
      
    W.append(w_i)
  
  weightsfile.close()
  
  print 'All importance weights saved to file weights.out'

  mean = sum(W) / (len(W) - 1)
  print 'Mean weight:', mean
  var = sum([(val - mean) ** 2 for val in W]) / (len(W) - 1) / len(W)
  print 'Standard error of the mean:', var ** 0.5

  